@@ -1,6 +1,6 @@
 # -*- coding: ISO-8859-15 -*-
 # =============================================================================
-# Copyright (c) 2008 Tom Kralidis
+# Copyright (c) 2024 Tom Kralidis
 #
 # Authors : Tom Kralidis <tomkralidis@gmail.com>
 #
@@ -11,8 +11,7 @@ import os
 import sys
 from collections import OrderedDict
 from dateutil import parser
-from datetime import datetime, timedelta
-import pytz
+from datetime import datetime, timedelta, timezone
 from owslib.etree import etree, ParseError
 from owslib.namespaces import Namespaces
 from urllib.parse import urlsplit, urlencode, urlparse, parse_qs, urlunparse, parse_qsl
@@ -24,6 +23,7 @@ import re
 from copy import deepcopy
 import warnings
 import requests
+from requests.auth import AuthBase
 import codecs
 
 """
@@ -173,8 +173,12 @@ def openURL(url_base, data=None, method='Get', cookies=None, username=None, pass
         verify = verify and auth.verify
     else:
         auth = Authentication(username, password, cert, verify)
+
     if auth.username and auth.password:
         rkwargs['auth'] = (auth.username, auth.password)
+    elif auth.auth_delegate is not None:
+        rkwargs['auth'] = auth.auth_delegate
+
     rkwargs['cert'] = auth.cert
     rkwargs['verify'] = verify
 
@@ -202,7 +206,7 @@ def openURL(url_base, data=None, method='Get', cookies=None, username=None, pass
 
     req = requests.request(method.upper(), url_base, headers=headers, **rkwargs)
 
-    if req.status_code in [400, 401]:
+    if req.status_code in [400, 401, 403]:
         raise ServiceException(req.text)
 
     if req.status_code in [404, 500, 502, 503, 504]:    # add more if needed
@@ -272,11 +276,8 @@ def nspath_eval(xpath, namespaces):
 
 def cleanup_namespaces(element):
     """ Remove unused namespaces from an element """
-    if etree.__name__ == 'lxml.etree':
-        etree.cleanup_namespaces(element)
-        return element
-    else:
-        return etree.fromstring(etree.tostring(element))
+    etree.cleanup_namespaces(element)
+    return element
 
 
 def add_namespaces(root, ns_keys):
@@ -287,35 +288,21 @@ def add_namespaces(root, ns_keys):
 
     ns_keys = [(x, namespaces.get_namespace(x)) for x in ns_keys]
 
-    if etree.__name__ != 'lxml.etree':
-        # We can just add more namespaces when not using lxml.
-        # We can't re-add an existing namespaces.  Get a list of current
-        # namespaces in use
-        existing_namespaces = set()
-        for elem in root.getiterator():
-            if elem.tag[0] == "{":
-                uri, tag = elem.tag[1:].split("}")
-                existing_namespaces.add(namespaces.get_namespace_from_url(uri))
-        for key, link in ns_keys:
-            if link is not None and key not in existing_namespaces:
-                root.set("xmlns:%s" % key, link)
-        return root
-    else:
-        # lxml does not support setting xmlns attributes
-        # Update the elements nsmap with new namespaces
-        new_map = root.nsmap
-        for key, link in ns_keys:
-            if link is not None:
-                new_map[key] = link
-        # Recreate the root element with updated nsmap
-        new_root = etree.Element(root.tag, nsmap=new_map)
-        # Carry over attributes
-        for a, v in list(root.items()):
-            new_root.set(a, v)
-        # Carry over children
-        for child in root:
-            new_root.append(deepcopy(child))
-        return new_root
+    # lxml does not support setting xmlns attributes
+    # Update the elements nsmap with new namespaces
+    new_map = root.nsmap
+    for key, link in ns_keys:
+        if link is not None:
+            new_map[key] = link
+    # Recreate the root element with updated nsmap
+    new_root = etree.Element(root.tag, nsmap=new_map)
+    # Carry over attributes
+    for a, v in list(root.items()):
+        new_root.set(a, v)
+    # Carry over children
+    for child in root:
+        new_root.append(deepcopy(child))
+    return new_root
 
 
 def getXMLInteger(elem, tag):
@@ -338,6 +325,43 @@ def getXMLInteger(elem, tag):
     return int(e.text.strip())
 
 
+def getXMLTree(rsp: ResponseWrapper) -> etree:
+    """
+    Parse a response into an XML elementtree instance
+    and raise a ValueError if the server returns a
+    non-XML response. The response may contain a useful
+    error message from the server.
+
+    Parameters
+    ----------
+
+    @param rsp: the ResponseWrapper for the XML request
+    """
+
+    raw_text = strip_bom(rsp.read())
+    et = etree.fromstring(raw_text)
+
+    # check for response type - if it is not xml then raise an error
+    # if the server doesn't provide a Content-Type then assume xml
+    content_type = rsp.info().get('Content-Type', 'text/xml')
+    url = rsp.geturl()
+
+    has_xml_tag = raw_text.lstrip().startswith(b'<?xml')
+
+    xml_types = ['text/xml', 'application/xml', 'application/vnd.ogc.wms_xml']
+    if not any(xt in content_type.lower() for xt in xml_types) and not has_xml_tag:
+        html_body = et.find('BODY')  # note this is case-sensitive
+        if html_body is not None and len(html_body.text) > 0:
+            response_text = html_body.text.strip("\n")
+        else:
+            response_text = raw_text
+
+        raise ValueError("%s responded with Content-Type '%s': '%s'" %
+                         (url, content_type, response_text))
+
+    return et
+
+
 def testXMLValue(val, attrib=False):
     """
 
@@ -379,7 +403,7 @@ def testXMLAttribute(element, attribute):
     return None
 
 
-def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, password=None, auth=None):
+def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, password=None, auth=None, headers=None):
     """
 
     Invoke an HTTP POST request
@@ -391,6 +415,8 @@ def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, p
     - request: the request message
     - lang: the language
     - timeout: timeout in seconds
+    - auth: owslib.util.Auth instance
+    - headers: HTTP headers to send with requests
 
     """
 
@@ -399,15 +425,22 @@ def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, p
 
     u = urlsplit(url)
 
-    headers = {
-        'User-Agent': 'OWSLib (https://geopython.github.io/OWSLib)',
+    headers_ = {
+        'User-Agent': 'OWSLib (https://owslib.readthedocs.io',
         'Content-type': 'text/xml',
-        'Accept': 'text/xml',
+        'Accept': 'text/xml,application/xml',
         'Accept-Language': lang,
         'Accept-Encoding': 'gzip,deflate',
         'Host': u.netloc,
     }
 
+    if headers:
+        headers_.update(headers)
+
+    if isinstance(request, dict):
+        headers_['Content-type'] = 'application/json'
+        headers_.pop('Accept')
+
     rkwargs = {}
 
     if auth:
@@ -419,14 +452,18 @@ def http_post(url=None, request=None, lang='en-US', timeout=10, username=None, p
         auth = Authentication(username, password)
     if auth.username is not None and auth.password is not None:
         rkwargs['auth'] = (auth.username, auth.password)
+    elif auth.auth_delegate is not None:
+        rkwargs['auth'] = auth.auth_delegate
     rkwargs['verify'] = auth.verify
     rkwargs['cert'] = auth.cert
 
-    up = requests.post(url, request, headers=headers, **rkwargs)
-    return up.content
+    if not isinstance(request, dict):
+        return requests.post(url, request, headers=headers_, **rkwargs)
+    else:
+        return requests.post(url, json=request, headers=headers_, **rkwargs)
 
 
-def http_get(*args, **kwargs):
+def http_prepare(*args, **kwargs):
     # Copy input kwargs so the dict can be modified
     rkwargs = copy.deepcopy(kwargs)
 
@@ -451,13 +488,39 @@ def http_get(*args, **kwargs):
     # Build keyword args for call to requests.get()
     if auth.username and auth.password:
         rkwargs.setdefault('auth', (auth.username, auth.password))
+    elif auth.auth_delegate is not None:
+        rkwargs['auth'] = auth.auth_delegate
     else:
         rkwargs.setdefault('auth', None)
     rkwargs.setdefault('cert', rkwargs.get('cert'))
     rkwargs.setdefault('verify', rkwargs.get('verify', True))
+
+    return rkwargs
+
+
+def http_get(*args, **kwargs):
+    rkwargs = http_prepare(*args, **kwargs)
     return requests.get(*args, **rkwargs)
 
 
+def http_put(*args, **kwargs):
+    rkwargs = http_prepare(*args, **kwargs)
+
+    if 'data' in kwargs:
+        if isinstance(kwargs['data'], dict):
+            rkwargs['json'] = kwargs['data']
+            rkwargs.pop('data')
+        else:
+            rkwargs['data'] = kwargs['data']
+
+    return requests.put(*args, **rkwargs)
+
+
+def http_delete(*args, **kwargs):
+    rkwargs = http_prepare(*args, **kwargs)
+    return requests.delete(*args, **rkwargs)
+
+
 def element_to_string(element, encoding=None, xml_declaration=False):
     """
     Returns a string from a XML object
@@ -475,21 +538,14 @@ def element_to_string(element, encoding=None, xml_declaration=False):
     if encoding is None:
         encoding = "ISO-8859-1"
 
-    if etree.__name__ == 'lxml.etree':
-        if xml_declaration:
-            if encoding in ['unicode', 'utf-8']:
-                output = '<?xml version="1.0" encoding="utf-8" standalone="no"?>\n{}'.format(
-                    etree.tostring(element, encoding='unicode'))
-            else:
-                output = etree.tostring(element, encoding=encoding, xml_declaration=True)
+    if xml_declaration:
+        if encoding in ['unicode', 'utf-8']:
+            output = '<?xml version="1.0" encoding="utf-8" standalone="no"?>\n{}'.format(
+                etree.tostring(element, encoding='unicode'))
         else:
-            output = etree.tostring(element)
+            output = etree.tostring(element, encoding=encoding, xml_declaration=True)
     else:
-        if xml_declaration:
-            output = '<?xml version="1.0" encoding="{}" standalone="no"?>\n{}'.format(
-                encoding, etree.tostring(element, encoding=encoding))
-        else:
-            output = etree.tostring(element)
+        output = etree.tostring(element)
 
     return output
 
@@ -546,7 +602,7 @@ def getNamespace(element):
         return ""
 
 
-def build_get_url(base_url, params, overwrite=False):
+def build_get_url(base_url, params, overwrite=False, doseq=False):
     ''' Utility function to build a full HTTP GET URL from the service base URL and a dictionary of HTTP parameters.
 
     TODO: handle parameters case-insensitive?
@@ -578,19 +634,23 @@ def build_get_url(base_url, params, overwrite=False):
         if key not in pars:
             qs.append((key, value))
 
-    urlqs = urlencode(tuple(qs))
+    urlqs = urlencode(tuple(qs), doseq=doseq)
     return base_url.split('?')[0] + '?' + urlqs
 
 
 def dump(obj, prefix=''):
     '''Utility function to print to standard output a generic object with all its attributes.'''
 
-    print(("{} {}.{} : {}".format(prefix, obj.__module__, obj.__class__.__name__, obj.__dict__)))
+    return "{} {}.{} : {}".format(prefix, obj.__module__, obj.__class__.__name__, obj.__dict__)
 
 
 def getTypedValue(data_type, value):
     '''Utility function to cast a string value to the appropriate XSD type. '''
 
+    # If the default value is empty
+    if value is None:
+        return
+
     if data_type == 'boolean':
         return True if value.lower() == 'true' else False
     elif data_type == 'integer':
@@ -624,8 +684,7 @@ Would be 2006-07-27T21:10:00Z, not 'now'
     except Exception:
         att = testXMLValue(element.attrib.get('indeterminatePosition'), True)
         if att and att == 'now':
-            dt = datetime.utcnow()
-            dt.replace(tzinfo=pytz.utc)
+            dt = datetime.utcnow().replace(tzinfo=timezone.utc)
         else:
             dt = None
     return dt
@@ -633,9 +692,12 @@ Would be 2006-07-27T21:10:00Z, not 'now'
 
 def extract_xml_list(elements):
     """
-Some people don't have seperate tags for their keywords and seperate them with
+Some people don't have separate tags for their keywords and separate them with
 a newline. This will extract out all of the keywords correctly.
 """
+    if elements is None:
+        return []
+
     keywords = (re.split(r'[\n\r]+', f.text) for f in elements if f.text)
     flattened = (item.strip() for sublist in keywords for item in sublist)
     remove_blank = [_f for _f in flattened if _f]
@@ -721,29 +783,6 @@ def bind_url(url):
     return '%s%s' % (url, binder)
 
 
-import logging
-# Null logging handler
-NullHandler = logging.NullHandler
-
-log = logging.getLogger('owslib')
-log.addHandler(NullHandler())
-
-
-def which_etree():
-    """decipher which etree library is being used by OWSLib"""
-
-    which_etree = None
-
-    if 'lxml' in etree.__file__:
-        which_etree = 'lxml.etree'
-    elif 'xml/etree' in etree.__file__:
-        which_etree = 'xml.etree'
-    elif 'elementree' in etree.__file__:
-        which_etree = 'elementtree.ElementTree'
-
-    return which_etree
-
-
 def findall(root, xpath, attribute_name=None, attribute_value=None):
     """Find elements recursively from given root element based on
     xpath and possibly given attribute
@@ -777,7 +816,7 @@ def datetime_from_iso(iso):
 
 def datetime_from_ansi(ansi):
     """Converts an ansiDate (expressed as a number = the nuber of days since the datum origin of ansi)
-    to a python datetime object.
+    to a Python datetime object.
     """
 
     datumOrigin = datetime(1600, 12, 31, 0, 0, 0)
@@ -833,11 +872,13 @@ class Authentication(object):
 
     _USERNAME = None
     _PASSWORD = None
+    _AUTH_DELEGATE = None
     _CERT = None
     _VERIFY = None
 
     def __init__(self, username=None, password=None,
-                 cert=None, verify=True, shared=False):
+                 cert=None, verify=True, shared=False,
+                 auth_delegate=None):
         '''
         :param str username=None: Username for basic authentication, None for
             unauthenticated access (or if using cert/verify)
@@ -852,12 +893,24 @@ class Authentication(object):
         :param bool shared=False: Set to True to make the values be class-level
             attributes (shared among instances where shared=True) instead of
             instance-level (shared=False, default)
+        :param AuthBase auth_delegate=None: Instance of requests' AuthBase to
+            allow arbitrary authentication schemes - mutually exclusive with
+            username/password arguments.
         '''
         self.shared = shared
+        self._username = username
+        self._password = password
+        self._cert = cert
+        self._verify = verify
+        self._auth_delegate = auth_delegate
+
+        # Trigger the setters to validate the parameters. This couldn't be done directly
+        # since some parameters are mutually exclusive.
         self.username = username
         self.password = password
         self.cert = cert
         self.verify = verify
+        self.auth_delegate = auth_delegate
 
     @property
     def username(self):
@@ -867,10 +920,15 @@ class Authentication(object):
 
     @username.setter
     def username(self, value):
-        if value is None:
-            pass
-        elif not isinstance(value, str):
-            raise TypeError('Value for "username" must be a str')
+        if value is not None:
+
+            if not isinstance(value, str):
+                raise TypeError('Value for "username" must be a str')
+
+            if self.auth_delegate is not None:
+                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
+                                 ' but not both')
+
         if self.shared:
             self.__class__._USERNAME = value
         else:
@@ -884,10 +942,15 @@ class Authentication(object):
 
     @password.setter
     def password(self, value):
-        if value is None:
-            pass
-        elif not isinstance(value, str):
-            raise TypeError('Value for "password" must be a str')
+        if value is not None:
+
+            if not isinstance(value, str):
+                raise TypeError('Value for "password" must be a str')
+
+            if self.auth_delegate is not None:
+                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
+                                 ' but not both')
+
         if self.shared:
             self.__class__._PASSWORD = value
         else:
@@ -947,8 +1010,32 @@ class Authentication(object):
         else:
             self._verify = value
 
+    @property
+    def auth_delegate(self):
+        if self.shared:
+            return self._AUTH_DELEGATE
+        return self._auth_delegate
+
+    @auth_delegate.setter
+    def auth_delegate(self, value):
+        if value is not None:
+            if not isinstance(value, AuthBase):
+                raise TypeError('Value for "auth_delegate" must be an instance of AuthBase')
+
+            if self.username is not None or self.password is not None:
+                raise ValueError('Authentication instances may have username/password or auth_delegate set,'
+                                 ' but not both')
+
+        if self.shared:
+            self.__class__._AUTH_DELEGATE = value
+        else:
+            self._auth_delegate = value
+
     @property
     def urlopen_kwargs(self):
+        if self.auth_delegate is not None:
+            raise NotImplementedError("The urlopen_kwargs property is not supported when auth_delegate is set")
+
         return {
             'username': self.username,
             'password': self.password,
@@ -957,5 +1044,6 @@ class Authentication(object):
         }
 
     def __repr__(self, *args, **kwargs):
-        return '<{} shared={} username={} password={} cert={} verify={}>'.format(
-            self.__class__.__name__, self.shared, self.username, self.password, self.cert, self.verify)
+        return '<{} shared={} username={} password={} cert={} verify={} auth_delegate={}>'.format(
+            self.__class__.__name__, self.shared, self.username, self.password, self.cert, self.verify,
+            self.auth_delegate)
